//
//   @file    applgrid_utils.cxx         
//   
//
//   @author M.Sutton
// 
//   Copyright (C) 2016 M.Sutton (sutt@cern.ch)    
//
//   $Id: applgrid_utils.cxx, v0.0   Mon 25 Apr 2016 15:03:34 CEST sutt $

#include "amconfig.h"

#ifdef HAVE_APPL

#include <iostream>
#include <set>

#include "appl_grid/appl_grid.h"
#include "appl_grid/appl_pdf.h"
#include "appl_grid/lumi_pdf.h"


#include "nnlo_common.h"
#include "nnlo_utils.h"


#include "applgrid_utils.h"

extern "C" void evolvepdf_(const double& x, const double& Q, double* xf);
extern "C" double alphaspdf_(const double& Q);


extern bool first_run;

//static std::int  _gridname; //  = "nnlojet.root";
static std::vector<appl::grid*>   _grids;
static std::map<int, appl::grid*> _gridmap;
static std::map<int, std::string> _gridnames;

namespace nnlo { 
void fill_appl_internal( appl::grid* _grid,
			 const int& order,
			 const double& x1, const double& x2, const double& Q2,
			 const double& obs, const double* wgt );
}


std::string findname( const int& id ) {
  std::map<int, std::string>::iterator gitr = _gridnames.find(id);
  if ( gitr!=_gridnames.end() ) return gitr->second;
  return "";
}



appl::grid* find( const int& id ) {
   std::map<int, appl::grid*>::iterator gitr = _gridmap.find(id);
   if ( gitr!=_gridmap.end() ) return gitr->second;
   std::cerr << "find() requested grid id " << id << " not found" << std::endl;
   return 0;
}







void nnlo::init_applgrid( const int& id, const std::string& gridname,  
			  int nbins, double lo, double hi ) {

  // From NNLOJETS common?
  int leading_order = inppar_.njets;
  int nloops        = 2;
  
  // Book APPLGRID
  
  std::cout << "### nnlo::init(): Book APPLGRID ..." << std::endl;
  
  // If file exists, APPLGRID warmup done already
  if ( file_exists( gridname ) ) {
    
    std::cout << "################### File exists ..." << std::endl;
    
    // Open the existing grid
    
    //    std::cout << "nnlo::init() ready " << _ready << std::endl;
    std::cout << "nnlo::init() reading grid " << gridname << " ..." << std::endl;

    appl::grid* _grid = new appl::grid( gridname);
    
    _gridnames.insert( std::map<int, std::string>::value_type( id,  gridname ) );
    _gridmap.insert(   std::map<int, appl::grid*>::value_type( id, _grid ) );
    
    _grid->optimise(true);
    
    first_run = false;
    
    std::cout << *_grid << std::endl;
    
  } else {
    
    std::cout << "nnlo::init() booking appl-grid from scratch with name " << gridname << " ..." << std::endl;
    
    /// Create a new grid - these should be set from some values
    /// from the NNLOJET code
    
    /// eta bins from eta0_a_gg5g0q_i1_n1_z2_q1_r001
    
    /// need to allow setting these from the fortran ???
    
    int    Nobs   = nbins;
    double obsmin = lo;
    double obsmax = hi;  //don't change this anymore !!!!!
    
    std::cout << "booking grid: " << Nobs << "\trange " << lo << " - " << hi << std::endl;
    
    int    NQ2     = 20;
    double Q2min   = 10;
    double Q2max   = 4*14000*14000;
    double Q2order = 4;
    
    int    Nx     = 40;
    double xmin   = 0.00001;
    double xmax   = 0.9999999;
    double xorder = 5;
    
    // int leading_order = inppar_.njets;
    // int nloops        = inppar_.nloop;
    
    std::cout << " njets: " <<  inppar_.njets << "\tnloops " << inppar_.nloop << std::endl;
    
    //    std::string genpdf = "ZJ-V.config";
    //    std::string genpdf = "ZJ-R.config";
    std::string genpdf = "ZJ-LO.config";
    
    std::cout << "creating grid " << genpdf << std::endl;
    
    //    lumi_pdf lumi( genpdf );
    //    lumi.write_latex( std::cout );
    
    //    double apramval=5.;
    //    appl::grid::transformvar(apramval);
    
    int nigrids = 0;
    
    if ( nloops == 0 ) nigrids = 0;
    if ( nloops == 1 ) nigrids = 1;
    if ( nloops == 2 ) nigrids = 2;
    
    appl::grid* _grid = new appl::grid( NQ2,  Q2min,  Q2max, Q2order,
					Nx,   xmin,   xmax,  xorder,
					Nobs, obsmin, obsmax,
					genpdf, leading_order, nigrids );
    
    // Use the reweighting function
    _grid->reweight(true);
    
    // Add documentation
    _grid->addDocumentation( banner() );
    
    _gridnames.insert( std::map<int, std::string>::value_type( id,  gridname ) );
    _gridmap.insert(   std::map<int, appl::grid*>::value_type( id, _grid ) );
    
  }
  
}


void nnlo::fill_applgrid( const int& id, 
			  const double& x1, const double& x2, const double& Q2, 
			  const double& obs, 
			  const double* wt ) 
{ 

  appl::grid* _grid = find( id );

  int iorder = ndimcurrent_.norder;

  if ( _grid ) {
    
    //         std::cout << "id:  " << id << "\nx1:  " << x1 << "\nx2:  " << x2 << "\nQ2: " << Q2 << std::endl;
    
    //         std::cout << "obs:  " << obs << std::endl;
    
    //        std::cout << "wt: \n";
    //    for ( int i=0 ; i<7 ; i++ ) std::cout << "\t" << wt[i] << std::endl;
    
    //    std::cout << "nnlo::fill() found grid " << _grid << std::endl;
    
    /// find iproc if need be
    
    const lumi_pdf* pdf = dynamic_cast<const lumi_pdf*>( ( (const appl::grid*)_grid)->genpdf(iorder) );
      
    int iproc = currentprocess_.iproc;
    
    /// get input parton species, calculate subprocess index
    
    /// Fixme: just for the time being - sort out which order we
    ///        are dealing with properly later
    
    
    /// use common/partid/ rather than common/idparton/ ??
    //    int  ip1 = partid_.ip1;
    //    int  ip2 = partid_.ip2;
    //    std::cout << "nnlo::fill() pdf " << pdf << "\tpartons " << ip1 << " " << ip2 << std::endl;
    
    int iprocess = pdf->decideSubProcess( iproc );
    
    //         std::cout << "nnlo::fill() iproc " << iproc << " " << iprocess << std::endl;
    
    if ( iprocess!=-1 ) {
      
      std::vector<double> wgt(pdf->size(), 0);
      
      wgt[iprocess] = wt[0];
      
      //      std::vector<double> weights(7,0);
      
      //      for ( int j=7 ; j-- ; ) weights[1] = wt[j];
      
      /// NB: now pass the weights in as a parameter ...
      /// chose which os the maxscl=3 weights, you actuall
      /// want to use
      //      weights[iprocess] = gridweights_.gridwgt[0];
      
      //      std::cout << "nnlo::fill() x1 " << x1 << "\tx2 " << x2 << "\tQ2 " << Q2 << std::endl;
      
      //      fill_appl_internal( _grid, iorder, x1, x2, Q2, obs, &weights[0] );
      fill_appl_internal( _grid, iorder, x1, x2, Q2, obs, &wgt[0] );
      
    }
    
  }
   
}


void nnlo::fill_appl_internal( appl::grid* _grid,
			       const int& order,
			       const double& x1, const double& x2, const double& Q2,
			       const double& obs, const double* wgt ) {

  //  static int Nsubproc = _grid->subProcesses( inppar_.nloop );

  const lumi_pdf* pdf = dynamic_cast<const lumi_pdf*>(((const appl::grid*)_grid)->genpdf( ndimcurrent_.norder ) );

  static std::vector<double> weight(pdf->size());

  static double  factor[4] = { 1, 1, 1, 1 };
  static bool   bfactor[4] = { false, false, false, false };

  //  int as_order = inppar_.njets + inppar_.nloop;
  int as_order = inppar_.njets + order;

  double alphas = alphaspdf_( std::sqrt(Q2) );

  if ( !bfactor[order] ) {
    bfactor[order] = true;
    for ( int iorder=0 ; iorder<as_order ; iorder++ ) factor[order] *= 2*M_PI/alphas;
  }

  int nv = ndimcurrent_.nv;

  if ( nv==0 ) {

    // tree level

    //  weight[0] = histonorm_.wthist * wt[0] * xjacob_.xjac1 / xi1 / xi2;
    //    for ( unsigned i=0 ; i<pdf->size() ; i++ ) weight[i] = wgt[i] / x1 / x2;
    for ( unsigned i=0 ; i<pdf->size() ; i++ ) weight[i] = wgt[i];

    for ( unsigned i=0 ; i<pdf->size() ; i++ ) weight[i] *= factor[order];

    if ( first_run ) _grid->fill_phasespace(  x1, x2, Q2, obs,  &weight[0], order );
    else             _grid->fill_grid(        x1, x2, Q2, obs,  &weight[0], order );

  }
  else if ( nv==2 ) {

    /// virtual term

    /// FIXME: move all these calculation of x1, x2 etc in to the
    /// calling function, to calculate once and for all for appl and fnlo,
    /// and then pass in the "scaled" x1 and x2 vales

    int ix = gridix_.igx; /// get this from our new, added common block

    double _x1 = x1;
    double _x2 = x2;

    if      ( ix==2 ) _x2 = x2/xregions_.xreg2;
    else if ( ix==3 ) _x1 = x1/xregions_.xreg1;
    else if ( ix==4 ) {
      _x1 = x1/xregions_.xreg1;
      _x2 = x2/xregions_.xreg2;
    }

    double jac = 1;

    if ( jacobian_.jflag ) jac = 1/((1-x1)*(1-x2));  /// or 1/((1-_x1)(1-_x2)) ???

    //  weight[0] = histonorm_.wthist * wt[0] * xjacob_.xjac1 / xi1 / xi2;
    //   for ( unsigned i=0 ; i<pdf->size() ; i++ ) weight[i] = wgt[i] / _x1 / _x2;
    for ( unsigned i=0 ; i<pdf->size() ; i++ ) weight[i] = wgt[i];

    for ( unsigned i=0 ; i<pdf->size() ; i++ ) weight[i] *= factor[order]*jac;

    if ( first_run ) _grid->fill_phasespace(  _x1, _x2, Q2, obs,  &weight[0], order );
    else             _grid->fill_grid(        _x1, _x2, Q2, obs,  &weight[0], order );

  }

}



void nnlo::fill_ref_applgrid( const int& id, const double& obs, const double* wt ) {
  appl::grid* _grid = find( id );
  if ( _grid ) _grid->getReference()->Fill( obs, wt[0] );
}








void nnlo::term_applgrid( const int& id, int nweights ) { 
  
  // Write APPLGRID
 
  static int icount = 0;
 
  std::cout << "### nnlo::term(): Write APPLGRID ... " << icount << std::endl;
  
  icount++;

  appl::grid* _grid = find( id );
  
  if ( _grid==0 ) return;
  

  std::string gridname = findname( id );
  
  if ( gridname=="" ) return;


  /// keep track of which rids we've already called 
  /// termination on  
  static std::set<int> term_id;
  if ( term_id.find(id)!=term_id.end() ) return;
  term_id.insert( id );


  /// we need to scale by the bin width ???
  
  //  double nweights = icount;
  
  ///  _grids.back()->getReference()->Scale(1/(conv*binwidth[nh]/nweights));
  
  _grid->run() = nweights;
  
  /// hmmm, we divide by the bin width after the convolution
  /// if the weights are *already* divided by the bin width, we will
  /// need to scale *up* the weights in each bin by the bin width
  /// as a hack ( for fixed binning ) do that here ...
  
  //      TH1D* h = _grid->getReference();
  
  
  /// scale up the grid by the number of weights - vegas already
  /// includes a 1/nshots term for the integration
  /// the grid will always be stored "unweighted"
  (*_grid) *= nweights;
  
  /// scale up the reference histogram, since it already is correctly
  /// weighted after the filling, but we scale it down by the number
  /// of weights when we write it out, so that the histogram in the
  /// root file is the actual cross section
  ///  _grids.back()->getReference()->Scale( nweights );
  
  Normalise( _grid->getReference() );
  
  _grid->Write( gridname );
  
}


#endif
